#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>

#ifndef HAVE_GETOPT_PROTOTYPE
int getopt(int argc, char * const *argv,  const  char  *optstring);
extern char *optarg;
extern int optind;
#endif

char * documentation = 
 " This program is designed to read in a set of exam scores and    \n"
 " generate an appropriate graphical representation. The problem   \n"
 " with displaying exam scores is that scores are an inaccurate    \n"
 " measure of student performance.  If they are displayed as delta \n"
 " spikes, it is difficult for the reader to judge the density of  \n"
 " exam scores in any given region; and there greater reason to    \n"
 " be concerned about loss of privacy; more fundamentally, the     \n"
 " display implies a level of accuracy that the exam does not      \n"
 " provide. A typical solution is to clump scores together by      \n"
 " ranges. This approach discards useful information about the     \n"
 " scores; for example, 80 and an 89, say, are treated as identical\n"
 " when decades are grouped.                                       \n"
 "                                                                 \n"
 " This program generates a plot of density of scores in units of  \n"
 " 'students per unit score'.  Each score is treated as a data item\n"
 " generated by a random variable (the student). Given the data    \n"
 " item, the distribution of this hypothesized random variable     \n"
 " is constructed. The distributions of each of the students are   \n"
 " summed to construct a hypothesized distribution for scores of   \n"
 " an infinitely-large student population. It is this distribution \n"
 " that is plotted.                                                \n"
 "                                                                 \n"
 " The main input parameter is the 'spread', which is an estimate  \n"
 " of the accuracy of the measurement performed by the test. If a  \n"
 " Gaussian distribution is assumed for the student's performance, \n"
 " the spread is the standard deviation of that performance. The   \n"
 " default form of the student random variable is binomial. The    \n"
 " fiction is that the test is composed of some number of          \n"
 " subproblems, and the student has a fixed probability of getting \n"
 " each of them right. The spread determines the number of         \n"
 " sub-problems so that for a score of 50, the distribution has    \n"
 " the specified standard deviation.                               \n"
 "                                                                 \n"
 " The student distribution is constructed according to aposteriori\n"
 " principles.  Given a data item D and a hypothesis H, the        \n"
 " probability of H given D is P(H|D) = P(D|H) * P(H)/P(D). The    \n"
 " term P(D|H) is given by the form of the hypothesized            \n"
 " distribution. The term P(H) is an apriori estimate of the       \n"
 " probabilities of various student distributions. The term P(D)   \n"
 " is determined by integrating all the terms P(D|H)*P(H) and      \n"
 " normalizing the integral to 1.  The default apriori distribution\n"
 " for H is p*(1-p) where p is the fractional score. This can      \n"
 " perhaps be justified on description-length principles; in       \n"
 " practice it gives the most pleasing plots. Other apriori        \n"
 " distributions, including the uniform distribution, are also     \n"
 " available.                                                      \n"
 "                                                                 \n"
 " Andrew Myers, March 2001                                        \n"
 "                                                                 \n"
 " May be used freely for non-profit                               \n"
 " purposes so long as this notice is maintained in the code.    \n\n"
 ;

#define BINS 500
#define MAXS 100000
#define DEFAULT_SPREAD 5
#define DEFAULT_MAX 100

/* Postscript parameters */
#define XOFF 50
#define YOFF 300
#define XSIZE 512
#define YSIZE 300
#define XSIZEFULL 640
#define YSIZEFULL 400
#define BORDER 50


#define PI 3.141592

#define FALSE 0
#define TRUE 1

int binomial_f = TRUE;
int do_weighting = TRUE;
int postscript = FALSE;
int full_axes = FALSE;

float grades[MAXS];
float curve[BINS];

float label_tick_ht = 5;
float major_tick_ht = 3.5;
float minor_tick_ht = 2;
float label_offset_x = 6;
float label_offset_y = 6;

float axis_line_width = 0.5;
float curve_width = 1;

int cumulative = FALSE;
int apriori_dist = 4;

char buf[20];

typedef float tick[2]; /* tick[0] is spacing, tick[1] is locn of first tick */
int do_average[2];

#define MAXBARS 100
int numbars = 0;
float vbars[MAXBARS];

#define MAXLABELS MAXBARS
int numlabels = 0;
float labelpos[MAXLABELS][2];
char *labels[MAXLABELS];

static int int_compare(const void *i, const void *j) {
    return *(int *)i - *(int *)j;
}

void usage(char *argv0)
{
    fprintf(stderr,"Usage: %s [-hnbwWpB] [-m <maxscore>] [-s <spread>]"
	    	   " [-v pos] [-t title] [-x <width>] [-y <height>] [file | - ]\n",
	    argv0);
    fprintf(stderr,"  -h : print a full description of the program\n");
    fprintf(stderr,"  -n : use gaussian distribution\n");
    fprintf(stderr,"  -b : use binomial distribution (default)\n");
    fprintf(stderr,"  -w : turn on weighting of student contributions (default)\n");
    fprintf(stderr,"  -W : turn off weighting of student contributions\n");
    fprintf(stderr,"  -p : PostScript output\n");
    fprintf(stderr,"  -m : set maximum score (default 100)\n");
    fprintf(stderr,"  -M : set minimum score (default 0)\n");
    fprintf(stderr,"  -B : draw axes on top and right as well\n");
    fprintf(stderr,"  -s : set spreading of scores (default %d)\n", DEFAULT_SPREAD);
    fprintf(stderr,"  -c : plot cumulative distribution\n");
    fprintf(stderr,"  -t : set title of graph for PostScript output\n");
    fprintf(stderr,"  -x <width>: set horizontal width (in.) for PostScript output\n");
    fprintf(stderr,"  -y <height>: set horizontal width (in.) for PostScript output\n");
    fprintf(stderr,"  -a<n>: set apriori distribution: \n"
"    0=uniform, 1=log DL, 2 = sqrt(sin), 3 = sin, 4 = DL, 5 = min(p,1-p),\n"
"    6=DL w/ p\n");
    fprintf(stderr,"  -A1 : show average\n");
    fprintf(stderr,"  -A2 : show median\n");
    fprintf(stderr,"  -v <x>: draw vertical divider at position x\n");
    fprintf(stderr,"  -L <label> <x> <y> : draw label at position (x,y)\n");
    exit(2);
}

float recroot2pi, logroot2pi;

float gaussian(float x1, float spread, float x2)
{
    float offset = (x1 - x2)*(1.0/spread);
    return (recroot2pi*(1.0/spread))* exp(-0.5 * offset * offset);
}

// approximate log(x!) to 3 digits (source: Knuth)
float log_factorial(float x)
{
    if (x < 1.0) return log_factorial(x+1.0) - log(x+1.0);
    else return (x + 0.5)*log(x) - x + (1.0/12.0)/x +
		(1.0/360.0)/(x*x*x) + logroot2pi;
}

/*
    Continuous binomial distribution:

    n/s = effitems
    i/n = x
*/
float cbinomial(float n, float x, float p)
{
    float on = x*n;
    float off = (1.0-x)*n;
    float sum = log_factorial(n) - log_factorial(on) - log_factorial(off);
	// exp(sum) = n choose on = n choose off
		
    if (p == 0.0) {
	if (on == 0.0)
	    sum += 0.0;
	else
	    return 0.0;
    } else
	sum += on * log(p);
    if (p == 1.0) {
	if (off == 0.0)
	    sum += 0.0;
	else
	    return 0.0;
    } else
	sum += off * log(1.0 - p);
    if (sum < -100.0) return 0.0;
    else return exp(sum);
}


/* The apriori distribution of student grades.  This is really just a
 * fudge factor to ensure that the plots look nice.  Forcing f(0) =
 * f(100) = 0 helps in this regard.  The non-fudged version (0) is just
 * to return 1.  The default is case 4, which can perhaps be justified
 * on MDL grounds. #5 can too, but the plots it produces aren't as nice!
 */
float apriori(float x, float min, float max) {
    float p = (x-min)/(max-min);
    switch (apriori_dist) {
	case 0: return 1;
	case 1: 
	    if (p == 0.0 || p == 1.0) return 0.0;
	    return -1.0/(log(p) + log(1.0-p));
	case 2:
	    return sin(M_PI * p);
	case 3:
	    return sqrt(sin(M_PI * p));
	case 4:
	    return p*(1.0-p);
	case 5:
        default:
	    return (p < 0.5) ? p : (1.0 - p);
    }
}

void moveto(float x, float y) {
    if (postscript) {
	printf("%.1f %.1f moveto\n", x, y);
    } else {
	printf("%.1f %.1f\n", x, y);
    }
}

void lineto(float x, float y) {
    if (postscript) {
	printf("%.3f %.3f lineto\n", x, y);
    } else {
	printf("%.3f %.3f\n", x, y);
    }
}

float xscale = 1.0;
float yscale = 1.0;
float xoff = 0.0;
float yoff = 0.0;
float x_0 = 0.0, y_0 = 0.0;
float xsize = XSIZE, ysize = YSIZE;

typedef struct {
    tick major, minor, label;
} axis;

void create_axis(float min, float max, axis/*OUT*/ *ax) {
    float biggest = (max-min)/3.0;
    float lg_major_x = log10(biggest);
    ax->label[0] = pow(10.0, floor(lg_major_x));

    if (ax->label[0]*5.0 < biggest) {
	ax->label[0] = ax->label[0] * 5;
	ax->major[0] = ax->label[0];
	ax->minor[0] = ax->label[0] / 5;
    } else
    if (ax->label[0]*2.0 < biggest) {
	ax->label[0] = ax->label[0] * 2;
	ax->major[0] = ax->label[0] / 2;
	ax->minor[0] = ax->label[0] / 10;
    } else {
	ax->major[0] = ax->label[0] / 2;
	ax->minor[0] = ax->label[0] / 10;
    }
    ax->label[1] = ax->label[0]*ceil(min/ax->label[0]);
    ax->major[1] = ax->major[0]*ceil(min/ax->major[0]);
    ax->minor[1] = ax->minor[0]*ceil(min/ax->minor[0]);
}

void draw_axis(int horizontal, const axis *ax, float pos, float min, float max,
	float label_offset) {
    int i;
    float x, y, v;
    float tx, ty;
    float lx = horizontal ? 0 : label_offset;
    float ly = horizontal ? label_offset : 0;
    assert(!isnan(min));
    assert(!isnan(max));
    for (i = 0;; i++) {
	float ht = minor_tick_ht;
	v = ax->minor[1] + i*ax->minor[0]; // DON'T STRENGTH-REDUCE THIS
	x = horizontal ? v : pos;
	y = horizontal ? pos : v;
	if (v > max) break;
	if (fabs(v/ax->major[0] - floor(0.5 + v/ax->major[0])) < 0.01) {
	    tx = horizontal ? 0 : major_tick_ht;
	    ty = horizontal ? major_tick_ht : 0;
	} else {
	    tx = horizontal ? 0 : minor_tick_ht;
	    ty = horizontal ? minor_tick_ht : 0;
	}
	if (fabs(v/ax->label[0] - floor(0.5 + v/ax->label[0])) < 0.01) {
	    float px, py;
	    ht = label_tick_ht;
	    px = x*xscale + xoff + lx;
	    py = y*yscale + yoff + ly;
	    printf("(%3g)\n", v);
	    printf("  dup stringwidth pop /w exch def\n");
	    printf("  dup false 0 0 moveto charpath flattenpath pathbbox\n"
                   "  /ury exch def /urx exch def /lly exch def /llx exch def\n");;
	    moveto(px, py);
	    if (horizontal) {
		printf("w 2 div neg ");
		if (label_offset >= 0) printf("0 ");
		else printf(" lly ury sub ");
	    } else {
		if (label_offset >= 0) printf("0 ");
		else printf("w neg ");
		printf("lly ury sub 2 div ");
	    }
	    printf("rmoveto show\n");
#if 0
	    if (horizontal) {
		printf(" dup stringwidth pop 2 div neg 0 rmoveto show\n");
	    } else {
		if (label_offset < 0) {
		     printf(
                    "(%3g) "
                    "dup stringwidth pop /w exch def "
                    "dup false 0 0 moveto charpath flattenpath pathbbox "
                    "/ury exch def /urx exch def /lly exch def /llx exch def "
                    "%.1f %.1f moveto "
                    "w neg lly ury sub 2 div rmoveto "
                    "show\n",
                    v, px, py);
		} else {
		     printf(
                    "(%3g) "
                    "dup false 0 0 moveto charpath flattenpath pathbbox "
                    "/ury exch def /urx exch def /lly exch def /llx exch def "
                    "%.1f %.1f moveto "
                    "0 lly ury sub 2 div rmoveto "
                    "show\n",
                    v, px, py);
		}
	    }
#endif
	}
        float x1 = x*xscale + xoff - tx, x2 = x*xscale + xoff + tx,
              y1 = y*yscale + yoff - ty, y2 = y*yscale + yoff + ty;
	printf("newpath %.1f %.1f moveto %.1f %.1f lineto stroke\n", x1, y1, x2, y2);
        assert(!isnan(x1)); assert(!isnan(x2));
        assert(!isnan(y1)); assert(!isnan(y2));
    }
    {
	float px0 = (horizontal ? min : pos) * xscale + xoff;
	float py0 = (horizontal ? pos : min) * yscale + yoff;
	float px1 = (horizontal ? max : pos) * xscale + xoff;
	float py1 = (horizontal ? pos : max) * yscale + yoff;
	printf("newpath\n");
	moveto(px0, py0);
	lineto(px1, py1);
	printf("stroke\n");
    }
}

int all_white_space(char *s) {
    for (; *s; s++) {
	if (!isspace(*s)) return 0;
    }
    return 1;
}

int main(int argc, char **argv)
{
    FILE *f;
    char *prog = argv[0];
    int ns = 0;
    int i,j;
    float spread = DEFAULT_SPREAD;
    int spread_opt = FALSE;
    float max = DEFAULT_MAX;
    float min = 0.0;
    float average, median;
    float ymax = 0.0;
    char *title = 0;
    int ch;
    int blank_lines = 0;
    recroot2pi = 1.0/sqrt(2.0*PI);
    logroot2pi = log(sqrt(2.0*PI));
    while (-1 != (ch = (getopt(argc, argv, "hbnwWpcBt:m:M:s:A:a:v:x:y:L:::")))) {
	switch (ch) {
	    case 'b':
		binomial_f = TRUE;
		break;
	    case 'n':
		binomial_f = FALSE;
		break;
	    case 'w':
		do_weighting = TRUE;
		break;
	    case 'W':
		do_weighting = FALSE;
		break;
	    case 'p':
		postscript = TRUE;
		break;
	    case 'c':
		cumulative = TRUE;
		break;
	    case 'm':
		max = atof(optarg);
		break;
	    case 'M':
		min = atof(optarg);
		break;
	    case 'B':
		full_axes = TRUE;
		break;
	    case 't':
		title = optarg;
		break;
	    case 'x':
		xsize = atof(optarg) * 72.0;
		break;
	    case 'y':
		ysize = atof(optarg) * 72.0;
		break;
	    case 's':
		spread = atof(optarg);
                spread_opt = TRUE;
		break;
	    case 'A': {
		int which = atoi(optarg)-1;
		if (which >= 0 || which < 2) 
		    do_average[which] = TRUE;
	    } break;
	    case 'v': {
		float x = atof(optarg);
		if (numbars < MAXBARS) vbars[numbars++] = x;
	    } break;
	    case 'L': {
		if (numlabels < MAXLABELS && optind+2 < argc) {
		    labels[numlabels] = optarg;
		    labelpos[numlabels][0] = atof(argv[optind+0]);
		    labelpos[numlabels][1] = atof(argv[optind+1]);
		    numlabels++;
		    optind = optind + 2;
		} else {
		    fprintf(stderr, "Illegal label specification\n");
		    usage(prog);
		    exit(1);
		} break;
	    } break;
	    case 'a':
		apriori_dist = atoi(optarg);
		break;
	    case 'h':
		puts(documentation);
		usage(prog);
		exit(0);
		break;
	    default:
		usage(prog);
	}
    }
    if (optind >= argc) usage(prog);

    if (optind < argc) {
	if (0 == strcmp(argv[optind], "-")) f = stdin;
	else {
	    f = fopen(argv[optind], "r");
	    if (!f) {
		perror(argv[optind]);
		usage(prog);
	    }
	}
    } else f = stdin;
    while (!feof(f)) {
	if (0 == fgets(buf, sizeof(buf), f)) break;
	if (all_white_space(buf)) {
	    blank_lines++;
	} else {
	    char *endptr;
	    grades[ns] = strtod(buf, &endptr);
	    if (endptr == buf) {
		fprintf(stderr, "Ignoring invalid number: %s\n", buf);
	    } else if (grades[ns] < min || grades[ns] > max) {
		fprintf(stderr, "Input out of range: %f\n", grades[ns]);
		exit(1);
	    } else {
		ns++;
	    }
	}
    }
    if (blank_lines > 0)
	fprintf(stderr, "Ignored %d blank line%s in input\n",
	    blank_lines, (blank_lines==1) ? "" : "s");
    if (do_average[1]) {
	qsort(grades, ns, sizeof(int), int_compare);
    }
    if (postscript) {
	printf("%%!PS-Adobe-2.0\n");
	printf("%%%%Creator: grader (written by Andrew Myers)\n");
	printf("%%%%Title: %s\n", title ? title :
		(argv[optind] ? argv[optind] : "stdin"));
	printf("%%%%Pages: 1\n");
	printf("%%%%PageOrder: Ascend\n");
	printf("%%%%BoundingBox: %d %d %d %d\n", XOFF-BORDER, YOFF-BORDER, XOFF+XSIZE+BORDER,
						YOFF+(int)ysize+BORDER);
	printf("%%%%DocumentFonts: Helvetica\n");
	printf("%%%%EndComments\n\n");

	printf("%%%%BeginSetup\n");
	printf("/Helvetica findfont 12 scalefont setfont\n");
	printf("%%%%EndSetup\n\n");

	printf("%%%%Page: 1 1\n");
	printf("false setstrokeadjust newpath\n");
    }
    float min_default_spread = (max-min)/30.0;
    if (!spread_opt && spread < min_default_spread) {
        spread = min_default_spread;
    }
    for (i=0; i<ns; i++) {
	float x = grades[i];
	float vec[BINS];
	float weight = 0.0;
	/* .48 is empirical constant... should calculate what it really is */
	float fispread = 0.48 * (max-min)/spread; /* fractional reciprocal spread */
	float effnumproblems = fispread * fispread;
	for (j = 0; j<BINS; j++) {
	    float frac = ((float)j/(BINS - 1));
	    float x1 = min + frac*(max - min);
	    if (cumulative) {
		if (grades[i] < x1) vec[j] = 1.0; else vec[j] = 0.0;
	    } else
	    if (binomial_f) {
		vec[j] = cbinomial(effnumproblems, (x-min)/(max-min), frac) *
		    		apriori(x1, min, max);
	    } else {
		vec[j] = gaussian(x, spread, x1)*apriori(x1, min, max);
	    }
	    weight += vec[j];
	}
	/* weight now contains integral of function being added to
	   curve, more or less. Now normalize and add to curve. */
	{
	    float iweight = 1.0;
	    if (do_weighting)
		iweight = BINS/(max - min)/weight;
	    for (j=0; j<BINS; j++) {
		curve[j] += vec[j] * iweight;
	    }
	}
    }
    for (j=0; j<BINS; j++) if (curve[j] > ymax) ymax = curve[j];

    if (cumulative) ymax = ns;

    if (postscript) {
	xscale = xsize/(max-min);
	x_0 = XOFF;
	xoff  = XOFF - min*xscale;
	yscale = ysize/ymax;
	y_0 = yoff = YOFF;
    }

    if (cumulative) {
	float x = 0.0;
	for (j = 1; j < ns; j++) {
	    int i = j - 1;
	    x = grades[j];
	    while (i >= 0 && grades[i] > x) {
		grades[i+1] = grades[i];
		i--;
	    }
	    grades[i+1] = x;
	}
	moveto(x_0,y_0);
	x = 0;
	for (i = 0; i < ns; i++) {
	    x = grades[i];
	    lineto(x*xscale + xoff, (float)i * yscale + yoff);
	    lineto(x*xscale + xoff, (i+1.0) * yscale + yoff);
	}
	lineto(max*xscale + xoff, (float)ns*yscale + yoff);
    } else {
	for (j=0; j<BINS; j++) {
	    float x2 = min + ((float)j/(BINS-1))*(max - min);
	    if (j == 0) {
		moveto(x2*xscale + xoff, curve[j]*yscale + yoff);
	    } else {
		lineto(x2*xscale + xoff, curve[j]*yscale + yoff);
	    }
	}
    }

    if (postscript) {
	float x, y;
	int i;
	axis x_axis;
	axis y_axis;
	int bin;

	printf("0 setgray %.2f setlinewidth stroke\n", curve_width);
	printf("newpath 0 setgray %.2f setlinewidth\n", axis_line_width);

	create_axis(min, max, &x_axis);
	create_axis(0, ymax, &y_axis);
	draw_axis(TRUE, &x_axis, 0, min, max, -label_offset_y);
	draw_axis(FALSE, &y_axis, min, 0, ymax, -label_offset_x);
	if (full_axes) {
	    draw_axis(TRUE, &x_axis, ymax, min, max, label_offset_y);
	    draw_axis(FALSE, &y_axis, max, 0, ymax, label_offset_x);
	}

	if (title) {
	    printf("/Helvetica-Bold findfont 12 scalefont setfont\n");
	    printf("newpath %.1f %.1f moveto (%s) dup false charpath"
		   " pathbbox\n",
		    x_0 + xsize/2.0, YOFF+ysize+0.0, title);
	    printf("/ury exch def /urx exch def /lly exch def /llx exch def\n");
	    printf("%.1f llx urx sub 2 div add\n"
		   "  %.1f lly add ury sub moveto show\n",
		    x_0 + xsize/2.0, YOFF+ysize+0.0);
	}


	if (do_average[0] || do_average[1])
	    printf("/Helvetica findfont 9 scalefont setfont\n");
	if (do_average[0]) { /* average */
	    float sum = 0.;
	    float ax, ay;
	    for (i = 0; i < ns; i++) {
		sum += grades[i];
	    }
	    average = sum/ns;
	    bin = (int)(((average-min) / (max-min)) * BINS + 0.5);
	    if (bin >= BINS) bin = BINS-1;
	    ax = average*xscale + xoff;
	    ay = curve[bin]*yscale + yoff + 10.0;
	    //fprintf(stderr, "ay = %f\n", ay);
	    printf("[3] 0 setdash\n");
	    printf("newpath %.1f %.1f moveto %.1f %.1f lineto stroke\n",
		    ax, yoff, ax, ay);
	    printf("[] 0 setdash\n");
	    printf("newpath %.1f %.1f moveto (average) dup "
		   "false charpath flattenpath pathbbox\n", ax, ay);
	    printf("/ury exch def /urx exch def /lly exch def /llx exch def ");
	    printf("%.1f llx urx sub 2 div add\n"
		   "  %.1f ury lly sub add 2 add moveto show\n",
		    ax, ay);
	}
	if (do_average[1]) { /* median */
	    float median;
	    float height;
	    float ax, ay;
	    switch (ns%2) {
		case 0: median = (grades[(ns-2)/2] + grades[ns/2])/2.0; break;
		case 1: median = grades[(ns-1)/2]; break;
	    }
	    ax = median*xscale + xoff;
	    bin = (int)(((median-min) / (max-min)) * BINS + 0.5);
	    ay = curve[bin]*yscale + yoff;
	    printf("/Helvetica findfont 9 scalefont setfont\n");
	    printf("[3] 0 setdash\n");
	    printf("newpath %.1f %.1f moveto %.1f %.1f lineto stroke\n",
		   ax, yoff, ax, ay);
	    printf("newpath %.1f %.1f moveto %.1f %.1f lineto stroke\n",
		    ax, yoff, ax, ay);
	    printf("[] 0 setdash\n");
	    printf("newpath %.1f %.1f moveto (median) dup "
		   "false charpath flattenpath pathbbox\n", ax, ay);
	    printf("/ury exch def /urx exch def /lly exch def /llx exch def\n");
	    printf("%% average = %f, median = %f\n", average, median);
	    if (do_average[0] && average >= median) {
		printf("%.1f llx add urx sub 2.0 sub\n", ax);
	    } else {
		printf("%.1f 2.0 add\n", ax);
	    }
	    printf("  %.1f lly ury sub 2 div add moveto show\n",
		    ay - 36.0);
	}
	for (i = 0; i < numbars; i++) {
	    float bx, by;
	    int bin;
	    float y;
	    x = vbars[i];
	    bin = (int)(((x-min) / (max-min)) * BINS);
	    if (bin < 0) {
		y = curve[0];
	    } else if (bin >= BINS) {
		y = curve[BINS-1];
	    } else {
		float w = ((x - min) - bin * (max - min)/BINS)/(max - min);
		y = curve[bin]*(1.0-w) + curve[bin+1]*w;
	    }
	    bx = x*xscale + xoff; by = y*yscale + yoff;
	    printf("newpath %.1f %.1f moveto %.1f %.1f lineto stroke\n",
		    bx, yoff, bx, by);
	}
	for (i = 0; i < numlabels; i++) {
	    float lx = labelpos[i][0], ly = labelpos[i][1];
	    float ax = lx*xscale + xoff, ay = ly*yscale + yoff;
	    char *label = labels[i];
	    printf("%% %s at %f, %f\n", label, lx, ly);
	    printf("/Helvetica findfont 9 scalefont setfont\n");
	    printf("[] 0 setdash\n");
	    printf("newpath %.1f %.1f moveto\n"
		   "(%s) false charpath flattenpath pathbbox\n", ax, ay, label);
	    printf("/ury exch def /urx exch def /lly exch def /llx exch def\n");
	    printf("%.1f llx urx sub 2.0 div add\n", ax);
	    printf("%.1f lly ury sub 2.0 div add\n", ay);
	    printf("moveto (%s) show\n", label);
	}
	printf("showpage\n");
	printf("%%%% EOF\n");

    }
    exit(0);
}
